/*
 math.c
  source file for useful math routines in MyCSeries

  Version 0.4.1

  Created: December 26, 2005
  Revised: December 27, 2006

 * THIS SOFTWARE IS NOT COPYRIGHTED
 *
 *  This source code is offered for use in the public domain.  You may use,
 * modify or distribute it freely.
 *
 *  This code is distributed in the hope that it will be useful but WITHOUT
 * ANY WARRANTY.  ALL WARRANTIES, EXPRESS OR IMPLIED ARE HEREBY DISCLAIMED.
 *  This includes but is not limited to warranties of MERCHANTABILITY or
 * FITNESS FOR A PARTICULAR PURPOSE.

December 22, 2005
  isqrt64 will now always use the [UInt64] division at the end because
    "sqrt" isn't promising that the operand will be [calculated] with 64 bit
    mantissa precision.
Monday, December 26, 2005
  moved the universal math specific code here
Friday, December 29, 2005
  added "twelfth_root_two" to the mix.  This is a necessary constant
    when it comes to calculating the frequencies of musical notes.
Wednesday, January 18, 2006
  after a lot of work in optimizing the addition code in my
    "Infinite Dig" software (HyperXDB), I added kLogBaseTwoTen constant.
Thursday, May 25, 2006
  reinstated the golden ratio constant
Friday, November 25, 2006
  the isqrt functions should now be consistent with the 68k assembly version -
    (as in, no error!).  Turns out Newton's method isn't perfect.
  isqrt64 no longer depends on an FPU - ever.
*/

#include <cs_math.h>

/* ------------ globals */

#ifndef env68k
  const _cs_eval_ldouble        long_double_pi = kPi;
  const _cs_eval_ldouble        long_double_radical_two = kRadicalTwo;
  const _cs_eval_ldouble        long_double_twelfth_root_two = kTwelfthRootTwo;
  const _cs_eval_ldouble        long_double_log2_10 = kLogBaseTwoTen;
  const _cs_eval_ldouble        long_double_golden_ratio = kGoldenRatio;

  const _cs_eval_double         double_t_one = 1.0;
  const _cs_eval_double         double_t_pi = kPi;
  const _cs_eval_double         double_t_radical_two = kRadicalTwo;
  const _cs_eval_double         double_t_twelfth_root_two = kTwelfthRootTwo;
  const _cs_eval_double         double_t_log2_10 = kLogBaseTwoTen;
  const _cs_eval_double         double_t_golden_ratio = kGoldenRatio;
#endif

/* ------------ code */

#undef LZCount32
#undef VFix16Mul
#undef VFix16Div

int LZCount32(UInt32 x);
sfixed VFix16Mul(sfixed a, sfixed b);
sfixed VFix16Div(sfixed n, sfixed d);

#ifndef env68k /* isqrt functions are implemented in assembly on env68k */
//#ifdef __MRC__
//#define sqrt __fsqrt
//#endif

  /* square root half (16 bit) table */
  static const UInt16   srhtb[] =
  {
     0x8084, 0x8180, 0x827C, 0x8378, 0x8470, 0x8568, 0x865C, 0x8750,
     0x8840, 0x8930, 0x8A1C, 0x8B0C, 0x8BF4, 0x8CE0, 0x8DC8, 0x8EAC,
     0x8F94, 0x9074, 0x9158, 0x9238, 0x9318, 0x93F8, 0x94D4, 0x95B0,
     0x9688, 0x9760, 0x9838, 0x9910, 0x99E8, 0x9ABC, 0x9B8C, 0x9C60,
     0x9D30, 0x9E00, 0x9ED0, 0x9F9C, 0xA06C, 0xA138, 0xA200, 0xA2CC,
     0xA394, 0xA45C, 0xA524, 0xA5E8, 0xA6AC, 0xA770, 0xA834, 0xA8F8,
     0xA9B8, 0xAA78, 0xAB38, 0xABF8, 0xACB8, 0xAD74, 0xAE30, 0xAEEC,
     0xAFA8, 0xB060, 0xB11C, 0xB1D4, 0xB28C, 0xB344, 0xB3F8, 0xB4B0,
     0xB564, 0xB618, 0xB6CC, 0xB780, 0xB830, 0xB8E4, 0xB994, 0xBA44,
     0xBAF4, 0xBBA0, 0xBC50, 0xBCFC, 0xBDAC, 0xBE58, 0xBF04, 0xBFB0,
     0xC058, 0xC104, 0xC1AC, 0xC254, 0xC300, 0xC3A4, 0xC44C, 0xC4F4,
     0xC598, 0xC640, 0xC6E4, 0xC788, 0xC82C, 0xC8D0, 0xC974, 0xCA14,
     0xCAB8, 0xCB58, 0xCBF8, 0xCC9C, 0xCD3C, 0xCDD8, 0xCE78, 0xCF18,
     0xCFB4, 0xD054, 0xD0F0, 0xD18C, 0xD228, 0xD2C4, 0xD360, 0xD3FC,
     0xD494, 0xD530, 0xD5C8, 0xD660, 0xD6F8, 0xD794, 0xD828, 0xD8C0,
     0xD958, 0xD9F0, 0xDA84, 0xDB1C, 0xDBB0, 0xDC44, 0xDCD8, 0xDD6C,
     0xDE00, 0xDE94, 0xDF28, 0xDFBC, 0xE04C, 0xE0E0, 0xE170, 0xE200,
     0xE294, 0xE324, 0xE3B4, 0xE444, 0xE4D4, 0xE560, 0xE5F0, 0xE680,
     0xE70C, 0xE798, 0xE828, 0xE8B4, 0xE940, 0xE9CC, 0xEA58, 0xEAE4,
     0xEB70, 0xEBFC, 0xEC88, 0xED10, 0xED9C, 0xEE24, 0xEEAC, 0xEF38,
     0xEFC0, 0xF048, 0xF0D0, 0xF158, 0xF1E0, 0xF268, 0xF2F0, 0xF374,
     0xF3FC, 0xF480, 0xF508, 0xF58C, 0xF614, 0xF698, 0xF71C, 0xF7A0,
     0xF824, 0xF8A8, 0xF92C, 0xF9B0, 0xFA34, 0xFAB4, 0xFB38, 0xFBBC,
     0xFC3C, 0xFCC0, 0xFD40, 0xFDC0, 0xFE44, 0xFEC4, 0xFF44, 0xFFFF,
  };

  UInt32 isqrt64(UInt64 n)
  {
     UInt64         t;

     /* to do this effectively, we get the highest ordered bit, and shift up... */
     unsigned long    a = Int64Hi(n), b = Int64Lo(n);
     unsigned long    r, s = LZCount32(a)&-2;

     if (a==0)
        return isqrt(b);

     a = (a<<s)|(b>>=32-s);
     r = srhtb[(a>>24)-64];

     /* ...and shift it back down by half when we are finished */
     s >>= 1;

     r += a/r; r >>= 1;

     /* ...and put it back to the right scale */
     r <<= 16-s;

     /* the table is large enough that we only have to divide and average twice */
     t = n;
     UDiv64U32(t,r);
     Add64U32(t,r);
     HalfU64(t);

     if (Int64Hi(t)!=0)
        r = UINT32_MAX; /* overflow */
     else
        r = Int64Lo(t);

     /* to prevent pre-rounding, we subtract one and test for rounding twice */
     UMul32(t,r,r-1);
     USub64(t,n);
     if (!Int64N(t))
        --r;
     if (r==UINT32_MAX)
        return UINT32_MAX;
     UMul32(t,r,r+1);
     USub64(t,n);
     if (Int64N(t))
        ++r;

     return r;
  }

  UInt32 isqrt(UInt32 n)
  {
     /* to do this effectively, we get the highest ordered bit, and shift up... */
     unsigned long    r, s = LZCount32(n)&-2;
     unsigned long    a = n<<s;

     if (n==0)
        return 0;
     r = srhtb[(a>>24)-64];

     /* ...and shift it back down by half when we are finished */
     s >>= 1;

     /* the table is large enough that we only have to divide and average once */
     r += a/r; r >>= 1;

     /* ...and scale it back down */
     r >>= s;

     /* to prevent pre-rounding, we subtract one and test for rounding twice */
     --r;
     if (r*r+r<n)
        ++r;
     if ((signed)(r*r+r-n)<0)
        ++r;

     return r;
  }
#endif /* !defined(env68k) */

/* prime divisibility table ## prime -t 2 3 5 7 */
static const UInt8      pdtb[] =
{
    2, 4, 2, 4,  6, 2, 6, 4,  2, 4, 6, 6,  2, 6, 4, 2,
    6, 4, 6, 8,  4, 2, 4, 2,  4, 8, 6, 4,  6, 2, 4, 6,
    2, 6, 6, 4,  2, 4, 6, 2,  6, 4, 2, 4,  2,10, 2,10
};

UInt32 *cs_prime_list(UInt64 n)
{
   static UInt32    r[65];

   UInt64           t;
   const UInt8      *s;
   UInt32           *d = r, i = 65, m, e, x, h;

   do
   {
      if ((e=isqrt64(n))<=1)
         break;

      if (!(Int64Lo(n)&1))
      {
         HalfU64(n);
         *d++ = 2;
         continue;
      }
      t = n;
      UDiv64U32R(t,3,m);
      if (m==0)
         {*d++ = 3; n = t; continue;}
      t = n;
      UDiv64U32R(t,5,m);
      if (m==0)
         {*d++ = 5; n = t; continue;}
      t = n;
      UDiv64U32R(t,7,m);
      if (m==0)
         {*d++ = 7; n = t; continue;}
      t = n, h = 11;
      UDiv64U32R(t,h,m);
      if (m==0)
         {*d++ = h; n = t; continue;}

      while(U64LeU32(n,e))
      {
         s = pdtb; x = GET_ARRAY_COUNT(pdtb);
         do
         {
            t = n, h += *s++;
            UDiv64U32R(t,h,m);
            if (m==0)
               {*d++ = h; n = t; break;}
         } while(--x);
      }
   } while(--i); /* mathematically speaking, it shouldn't loop 65 times */

   *d = 0;

   return r;
}

/* these two functions are merely glue */
UInt32 _isqrt64(UInt32 a, UInt32 b)
{
   UInt64           t;

   SetInt64(t,a,b);

   return isqrt64(t);
}

ufixed FixedSqrt(ufixed x)
{
   UInt64           t;

   SetInt64(t,x>>16,x<<16);

   return isqrt64(t);
}

/* to avoid any linker problems... */

int LZCount32(UInt32 x)
{
   int              i = 32;

   do{if (x&0x80000000) break; x += x;} while(--i);

   return 32-i;
}

#ifndef env68k
SInt32 ClippedS64ToS32(SInt64 x)
{
   if (S64GtU32(x,SINT32_MAX))
      return SINT32_MAX;
   else if (Int64Hi(x)<=(-1) && (Int64Hi(x)<(-1) || ((signed)Int64Lo(x))>=0))
      return SINT32_MIN;

   return Int64Lo(x);
}

sfixed VFix16Mul(sfixed a, sfixed b)
{
   SInt64           t;

   SMul32(t,a,b);
   if (Int64N(t))
   {
      if (Int64Hi(t)<=SINT16_MIN)
         return SINT32_MIN;
      Add64U32(t,SINT16_MAX);
   } else
   {
      Add64U32(t,SINT16_MAX+1);
      if (Int64Hi(t)>=SINT16_MAX)
         return SINT32_MAX;
   }
   RShiftS64(t,16);

   return Int64Lo(t);
}

sfixed VFix16Div(sfixed n, sfixed d)
{
   double           t;

   if (d==0)
      return n>=0 ? SINT32_MAX : SINT32_MIN;

   /* test for overflow before rounding to reduce the chance for error */
   if ((t=(n*65536.0)/d)<0)
   {
      if (t<=SINT32_MIN)
         return SINT32_MIN;
      t -= 0.5;
   } else
   {
      if (t>=SINT32_MAX)
         return SINT32_MAX;
      t += 0.5;
   }

   return (sfixed)t;
}
#endif
